# -*- coding: utf-8 -*-

import logging
from pathlib import Path

import yaml

logger = logging.getLogger(__name__)


def recursive_update(original_dict: dict, new_dict: dict) -> dict:
    """Recursively update original_dict with new_dict"""
    for new_key, new_value in new_dict.items():
        if isinstance(new_value, dict):
            original_dict[new_key] = recursive_update(
                original_dict.get(new_key, {}), new_value
            )
        else:
            original_dict[new_key] = new_value
    return original_dict


class Password:
    def __init__(self, password: str) -> None:
        self.password = password or ""

    def __repr__(self) -> str:
        return "*" * len(self.password)

    def get(self) -> str:
        return self.password

    def __bool__(self):
        return bool(self.password)


# Configurations are loaded from the defaults of the package and eventually a local config.yaml file
config_files = [
    Path(__file__).parent / "resources" / "default_config.yaml",
    Path("config.yaml"),
]

config = {}
for config_file in config_files:
    if config_file.exists():
        new_config = yaml.safe_load(config_file.read_text())
        if isinstance(new_config, dict):
            config = recursive_update(config, new_config)


config["backup-dir"] = Path(config["backup-dir"]).absolute()
config["password"] = Password(config.get("password"))



#! /usr/bin/env python

from pathlib import Path

import attr
import pendulum

from .client import GarminClient


@attr.s(frozen=True)
class Wellness:
    """Garmin wellness data


    Parameters
    ----------
    date
        Date for which you want to fetch wellness data

    """

    date: pendulum.DateTime = attr.ib()

    @date.validator
    def wont_travel_to_future(self, attribute, value):
        if value.in_tz("local") > pendulum.DateTime.today().in_tz("local"):
            raise ValueError(
                f"garpy cannot download data from the future... "
                f"try a date before today {value.format('YYYY-MM-DD')}"
            )

    @property
    def base_filename(self) -> str:
        return self.date.format("YYYY-MM-DD") + ".zip"

    def get_export_filepath(self, backup_dir: Path) -> Path:
        return Path(backup_dir) / self.base_filename

    def download(self, client: GarminClient, backup_dir: Path) -> None:
        """Download activity on the given format to the given backup directory

        Parameters
        ----------
        client
            Authenticated GarminClient
        backup_dir
            Where to download the file
        """
        response = client.get_wellness(self.date)
        backup_dir = Path(backup_dir)
        backup_dir.mkdir(exist_ok=True)
        filepath = self.get_export_filepath(backup_dir)
        if response.status_code == 200:
            filepath.write_bytes(response.content)
        else:
            with open(str(Path(backup_dir) / ".not_found"), mode="a") as not_found:
                not_found.write(str(filepath.name) + "\n")



#! /usr/bin/env python

import logging
from pathlib import Path
from typing import Dict

import attr

from . import Activities, Activity, GarminClient
from .settings import config

logger = logging.getLogger(__name__)

DEFAULT_FORMATS = tuple(config["activities"].keys())


@attr.s
class ActivitiesDownloader:
    """Class for doing incremental backups of your Garmin activitiess

    Parameters
    ----------
    client
        Authenticated GarminClient
    backup_dir
        Where to download the file
    """

    client: GarminClient = attr.ib()
    backup_dir: Path = attr.ib(default=config["backup-dir"])

    @backup_dir.validator
    def enforce_path(self, attribute, value):
        """Make sure that self.backup_dir is cast into Path and that it exits"""
        if isinstance(value, str):
            self.backup_dir = Path(value)
        self.backup_dir = self.backup_dir.absolute()
        self.backup_dir.mkdir(exist_ok=True)

    @property
    def existing_files(self):
        """Set of existing files on the backup directory"""
        return set(filepath for filepath in self.backup_dir.glob("*"))

    @property
    def not_found(self):
        """Set of files not found on a previous backup try"""
        if (self.backup_dir / ".not_found").exists():
            return set(
                self.backup_dir / path.strip()
                for path in (self.backup_dir / ".not_found").read_text().split("\n")
                if path
            )
        else:
            return set()

    def _discover_formats_to_download(
        self, activities: Activities, formats: tuple = DEFAULT_FORMATS
    ) -> Dict[Activity, tuple]:
        """Fetch list of activities and find which of them need backup and in what formats"""

        files_not_to_download = self.existing_files | self.not_found
        to_download = {}
        for activity in activities:
            needed_formats = tuple(
                fmt
                for fmt in formats
                if activity.get_export_filepath(self.backup_dir, fmt)
                not in files_not_to_download
            )
            if needed_formats:
                to_download[activity] = needed_formats

        return to_download

    def download_all(self, activities: Activities, formats: tuple = DEFAULT_FORMATS):
        """Do an incremental backup of the specified formats.

        Parameters
        ----------
        activities
            List of activities fetched from Garmin Connect
        formats
            Formats you wish to download
        """
        to_download = self._discover_formats_to_download(activities, formats)
        if not to_download:
            logger.info("Backup folder up to date. No activities will be downloaded")
            return

        n_activities = len(to_download)
        logger.info(f"{n_activities} activities to be downloaded")

        to_download = progressbar(to_download.items())
        for i, (activity, formats) in enumerate(to_download):
            to_download.desc = (
                f"Downloading {activity.type!r} activity {activity.id!r} "
                f"from {activity.start.format('YYYY-MM-DD')}. Formats: {formats!r}"
            )
            to_download.display()
            formats = progressbar(formats, leave=(i + 1 == n_activities))
            for fmt in formats:
                formats.desc = f"Downloading format {fmt!r}"
                formats.display()
                activity.download(self.client, fmt, self.backup_dir)

    def download_one(self, activity: Activity, formats: tuple = DEFAULT_FORMATS):
        """Download specified formats for a given activity

        Parameters
        ----------
        activity
            Activity you wish to download
        formats
            Formats you wish to download
        """
        logger.info(
            f"Downloading {activity.type!r} activity {activity.id!r} "
            f"from {activity.start.format('YYYY-MM-DD')}."
        )
        formats = progressbar(formats, leave=True)
        for fmt in formats:
            formats.desc = f"Downloading format {fmt!r}"
            formats.display()
            activity.download(self.client, fmt, self.backup_dir)

    def __call__(self, formats: tuple = DEFAULT_FORMATS, activity_id: int = None):
        logger.info(
            f"Downloading the following formats: {formats!r} "
            f"to this folder: {self.backup_dir}"
        )
        if activity_id is None:
            logger.info("Querying list of activities")
            activities = Activities.list(self.client)
            logger.info(
                f"{len(activities)} activities in total found on Garmin Connect"
            )
            self.download_all(activities=activities, formats=formats)
        else:
            logger.info(f"Fetching summary information for activity: {activity_id!r}")
            activity = Activity.from_garmin_connect(activity_id, self.client)
            self.download_one(activity=activity, formats=formats)


def _isnotebook():  # pragma: no cover
    """Check if garpy is being run inside a Jupyter notebook"""
    try:
        shell = get_ipython().__class__.__name__
        return shell == "ZMQInteractiveShell"
    except NameError:
        return False  # Probably standard Python interpreter or IPython


def progressbar(*args, **kwargs):  # pragma: no cover
    """Make a progress bar depending on the environment garpy is running"""
    if _isnotebook():
        from tqdm import tqdm_notebook as tqdm
    else:
        from tqdm import tqdm

    return tqdm(*args, **kwargs)



#! /usr/bin/env python

import json
import os
import zipfile
from io import BytesIO
from pathlib import Path
from typing import Any, Dict

import attr
import pendulum

from .client import GarminClient
from .settings import config


@attr.s(frozen=True)
class Activity:
    """Garmin activity

    Parameters
    ----------
    id
        Activity ID on Garmin Connect
    name
        Name of the activity
    type
        Type of activity, e.g, Cycling, Swimming, etc
    start
        Start date and time

    Examples
    --------
    .. code-block:: python

        >>> with GarminClient("my.sample@sample.com", "secretpassword") as client:
        >>>     activity = Activity.from_garmin_connect(3983141717, client)

    """

    id: int = attr.ib()
    name: str = attr.ib()
    type: str = attr.ib()
    start: pendulum.DateTime = attr.ib()

    @property
    def base_filename(self):
        filename = f"{self.start.in_tz('UTC').isoformat()}_{self.id}"
        if os.name == "nt":  # Windows complains about : in filenames.
            filename = filename.replace(":", ".")
        return filename

    def get_export_filepath(self, backup_dir: Path, fmt: str) -> Path:
        format_parameters = config["activities"].get(fmt)
        if not format_parameters:
            raise ValueError(f"Format '{fmt}' unknown.")
        return Path(backup_dir) / (self.base_filename + format_parameters["suffix"])

    @classmethod
    def from_garmin_summary(cls, summary: Dict[str, Any]):
        """Constructor based on garmin connect summary.

        Parameters
        ----------
        summary
            JSON string representation of summary information fetched from garmin connect
        """

        return cls(
            id=summary["activityId"],
            name=summary["activityName"],
            type=summary["activityTypeDTO"]["typeKey"],
            start=pendulum.parse(
                summary["summaryDTO"]["startTimeLocal"],
                tz=summary["timeZoneUnitDTO"]["unitKey"],
            ),
        )

    @classmethod
    def from_garmin_activity_list_entry(cls, entry: Dict[str, Any]):
        """Constructor based on an entry from the list of activities from garmin connect.

        Parameters
        ----------
        entry
            JSON string representation of an entry of an activity list fetched from garmin connect
        """

        return cls(
            id=entry["activityId"],
            name=entry["activityName"],
            type=entry["activityType"]["typeKey"],
            # Unfortunately, Garmin connect does not provide timezone information on entries from list of activities
            start=pendulum.parse(entry["startTimeGMT"]),
        )

    @classmethod
    def from_garmin_connect(cls, activity_id: int, client: GarminClient):
        """Constructor that fetches activity summary from Garmin Connect

        Parameters
        ----------
        activity_id
            Activity ID on Garmin Connect
        client
            Authenticated GarminClient
        """
        response = client.get_activity(activity_id, "summary")
        activity_summary = json.loads(response.text)
        activity = Activity.from_garmin_summary(activity_summary)

        return activity

    def download(self, client: GarminClient, fmt: str, backup_dir: Path) -> None:
        """Download activity on the given format to the given backup directory

        Parameters
        ----------
        client
            Authenticated GarminClient
        fmt
            Format you wish to download
        backup_dir
            Where to download the file
        """
        response = client.get_activity(self.id, fmt)
        filepath = self.get_export_filepath(backup_dir, fmt)
        backup_dir = Path(backup_dir)
        backup_dir.mkdir(exist_ok=True)
        if response.status_code == 200:
            if fmt == "original":
                zip_content = zipfile.ZipFile(BytesIO(response.content), mode="r")
                original_file_name = zip_content.namelist()[0]
                fit_bytes = zip_content.open(original_file_name).read()
                # Change file extension to the one on the zipped file
                filepath = filepath.with_suffix(Path(original_file_name).suffix)
                filepath.write_bytes(fit_bytes)

                # If original format is not FIT, register it as not found
                if filepath.suffix != ".fit":
                    with open(
                        str(Path(backup_dir) / ".not_found"), mode="a"
                    ) as not_found:
                        not_found.write(
                            str(self.get_export_filepath(backup_dir, fmt).name) + "\n"
                        )
            else:
                filepath.write_text(response.text)
        else:
            with open(str(Path(backup_dir) / ".not_found"), mode="a") as not_found:
                not_found.write(str(filepath.name) + "\n")


class Activities(list):
    @classmethod
    def list(cls, client: GarminClient):
        return cls(
            Activity.from_garmin_activity_list_entry(activity)
            for activity in client.list_activities()
        )



__version__ = "0.3.8"

import logging
import sys

from .activity import Activities, Activity
from .client import GarminClient
from .download import ActivitiesDownloader
from .wellness import Wellness

# Create logger
logger = logging.getLogger(__name__)

# Avoid duplicate handlers
logger.handlers = []

# Create STDERR handler
handler = logging.StreamHandler(sys.stderr)
handler.setFormatter(
    logging.Formatter("%(asctime)s %(levelname)s - %(message)s", "%Y-%m-%d %H:%M:%S")
)
handler.setLevel(logging.INFO)
logger.addHandler(handler)

# Prevent multiple logging if called from other packages
logger.propagate = False
logger.setLevel(logging.INFO)



from pathlib import Path

import click

from garpy import ActivitiesDownloader, GarminClient
from garpy.settings import config

FORMATS = set(config.get("activities").keys()) | {"fit"}


@click.group()
@click.version_option()
def main():
    pass


@main.command()
@click.argument("backup-dir", default=config.get("backup-dir"))
@click.option(
    "--formats",
    "-f",
    multiple=True,
    help="Which formats to download. The flag can be used several times, e.g. '-f original -f gpx'",
    type=click.Choice(FORMATS),
    show_choices=True,
    default=FORMATS,
)
@click.option(
    "--username",
    "-u",
    prompt=True,
    default=config.get("username"),
    metavar="{username}",
    help="Username of your Garmin account",
)
@click.option(
    "--password",
    "-p",
    prompt=True,
    default=config.get("password").get(),
    metavar="{password}",
    help="Password of your Garmin account",
    hide_input=True,
)
@click.option(
    "--activity",
    "-a",
    "activity_id",
    default=None,
    metavar="{ID}",
    help="Activity ID. If indicated, download only that activity, even if it has already been downloaded."
    " Otherwise, do incremental update of backup",
    hide_input=True,
)
@click.option(
    "--user-agent",
    "user_agent",
    default=config["user-agent"],
    metavar="{user_agent}",
    help="User agent to be used by requests",
    hide_input=True,
)
def download(backup_dir, formats, username, password, activity_id, user_agent):
    """Download activities from Garmin Connect

    Entry point for downloading activities from Garmin Connect. By default, it downloads all
    newly created activities since the last time you did a backup.

    If you specify an activity ID with the "-a/--activity" flag, only that activity will be downloaded,
    even if it has already been downloaded before.

    If no format is specified, the app will download all possible formats. Otherwise you can specify the
    formats you wish to download with the "-f/--formats" flag. The flag can be used several  times if you
    wish to specify several formats, e.g., 'garpy download [OPTIONS] -f original -f gpx [BACKUP_DIR]' will
    download .fit and .gpx files
    """
    if "fit" in formats:
        formats = set(formats)
        formats.remove("fit")
        formats.add("original")
        formats = tuple(formats)

    backup_dir = Path(backup_dir).absolute()
    if backup_dir.is_file():
        raise Exception("The provided backup directory exists and is a file")

    with GarminClient(
        username=username, password=password, user_agent=user_agent
    ) as client:
        downloader = ActivitiesDownloader(client=client, backup_dir=backup_dir)
        downloader(formats=formats, activity_id=activity_id)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['garpy']

package_data = \
{'': ['*'], 'garpy': ['resources/*']}

install_requires = \
['PyYAML>=5.1,<7.0',
 'attrs>=21.4.0,<22.0.0',
 'click>=7,<9',
 'cloudscraper>=1.2.58,<2.0.0',
 'pendulum>=2.0,<3.0',
 'requests>=2.22,<3.0',
 'tqdm>=4.36,<5.0']

entry_points = \
{'console_scripts': ['garpy = garpy.cli:main']}

setup_kwargs = {
    'name': 'garpy',
    'version': '0.3.8',
    'description': 'Python client for downloading activities from Garmin Connect',
    'long_description': '###################################\nGarpy: Make your garmin data yours!\n###################################\n\n|PyPI-Versions| |PyPI-Status| |Codacy-Grade| |Tests| |Coveralls|\n\n``garpy`` is a simple app used to backup your data from Garmin Connect. It can be used to do incremental\nbackups of your data from Garmin Connect or to download one specific activity.\n\n********************************\nIncremental backup of activities\n********************************\n\nThe first time you use it, all the activities found on your Garmin Connect account will be downloaded to\nthe directory that you specify. Afterwards, each time you run the command, only the newly available\nactivities will be downloaded.\n\nThe command is used as follows:\n\n.. code:: sh\n\n    garpy download {backup-dir}\n\nBehind the scenes, this is what will happen:\n\n- `garpy` will prompt you for your password and will then authenticate against Garmin Connect.\n- It will first fetch the list of all your activities from garmin.\n- It will check which activities have already been backed up on the given `backup-dir`\n- It will proceed to download all the missing activities.\n\n************************************\nDownloading one activity from its ID\n************************************\n\nIf you wish to download only one activity or simple you want to refresh an already downloaded activity,\nuse the \'-a/--activity\' flag as follows:\n\n.. code:: sh\n\n    garpy download --activity 1674567326 {backup-dir}\n\nThis will download the activity in all existing formats to the given `backup_dir`\n\n****************\nFull CLI options\n****************\n\nFor more detailed usage, invoke the \'--help\' command:\n\n.. code:: sh\n\n    $ garpy download --help\n    Usage: garpy download [OPTIONS] [BACKUP_DIR]\n\n      Download activities from Garmin Connect\n\n      Entry point for downloading activities from Garmin Connect. By default, it\n      downloads all newly created activities since the last time you did a backup.\n\n      If you specify an activity ID with the "-a/--activity" flag, only that\n      activity will be downloaded, even if it has already been downloaded before.\n\n      If no format is specified, the app will download all possible formats.\n      Otherwise you can specify the formats you wish to download with the\n      "-f/--formats" flag. The flag can be used several  times if you wish to\n      specify several formats, e.g., \'garpy download [OPTIONS] -f original -f gpx\n      [BACKUP_DIR]\' will download .fit and .gpx files\n\n      Options:\n      -f, --formats [original|gpx|fit|tcx|kml|summary|details]\n                                      Which formats to download. The flag can be\n                                      used several times, e.g. \'-f original -f\n                                      gpx\'\n      -u, --username {username}       Username of your Garmin account\n      -p, --password {password}       Password of your Garmin account\n      -a, --activity {ID}             Activity ID. If indicated, download only\n                                      that activity, even if it has already been\n                                      downloaded. Otherwise, do incremental update\n                                      of backup\n      --user-agent {user_agent}       User agent to be used by requests\n      --help                          Show this message and exit.\n\n\n************\nInstallation\n************\n``garpy`` requires Python 3.7 or higher on your system. For those who know your way around with Python, install\n``garpy`` with pip as follows:\n\n.. code:: sh\n\n    pip install -U garpy\n\n\nIf you are new to Python or have Python 2 installed on your\ncomputer, I recommend you install Miniconda_. To my knowledge, it is the simplest way of installing a robust and\nlightweight Python environment.\n\n\n****************\nAcknowledgements\n****************\n\nThe library is based on garminexport_. I borrowed the GarminClient, refactored it to my taste and\ncreated a package from it.\n\n\n.. |PyPI-Versions| image:: https://img.shields.io/pypi/pyversions/garpy.svg?logo=python&logoColor=white\n   :target: https://pypi.org/project/garpy\n.. |PyPI-Status| image:: https://img.shields.io/pypi/v/garpy.svg\n   :target: https://pypi.org/project/garpy\n.. |Codacy-Grade| image:: https://api.codacy.com/project/badge/Grade/2fbbd268e0a04cd0983291227be53873\n   :target: https://app.codacy.com/manual/garpy/garpy/dashboard\n.. |Tests| image:: https://github.com/felipeam86/garpy/actions/workflows/test.yml/badge.svg\n    :target: https://github.com/felipeam86/garpy/actions/workflows/test.yml\n.. |Coveralls| image:: https://coveralls.io/repos/github/felipeam86/garpy/badge.svg?branch=develop\n    :target: https://coveralls.io/github/felipeam86/garpy?branch=develop\n\n\n.. _Miniconda: https://docs.conda.io/en/latest/miniconda.html\n.. _garminexport: https://github.com/petergardfjall/garminexport\n',
    'author': 'Felipe Aguirre Martinez',
    'author_email': 'felipeam86@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/felipeam86/garpy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.9,<4.0.0',
}


setup(**setup_kwargs)



#! /usr/bin/env python

"""
A module for authenticating against and communicating with selected parts of the Garmin Connect REST API.

# The client is originally inspired by:
https://github.com/petergardfjall/garminexport

# Other useful reference used by the original garminexport project:
https://github.com/cpfair/tapiriik/blob/master/tapiriik/services/GarminConnect/garminconnect.py
"""

import json
import logging
import re
import sys
from typing import Dict, List, Tuple

import attr
import cloudscraper
import pendulum
import requests

from .settings import Password, config

logger = logging.getLogger(__name__)
ENDPOINTS = config["endpoints"]


def extract_auth_ticket_url(auth_response: str):
    """Extract authentication ticket URL from response of authentication form submission.

    The auth ticket URL is typically of form:

    https://connect.garmin.com/modern?ticket=ST-0123456-aBCDefgh1iJkLmN5opQ9R-cas

    Parameters
    ----------
    auth_response :
        HTML response from an auth form submission.

    Returns
    -------
    str
        Authentication ticket url
    """

    match = re.search(r'response_url\s*=\s*"(https:[^"]+)"', auth_response)
    if not match:
        raise ConnectionError(
            "auth failure: unable to extract auth ticket URL. did you provide a correct username/password?"
        )
    auth_ticket_url = match.group(1).replace("\\", "")
    logger.debug("auth ticket url: '%s'", auth_ticket_url)
    return auth_ticket_url


@attr.s
class GarminClient(object):
    """A client class used to authenticate with Garmin Connect

    Since this class implements the context manager protocol, this object
    can preferably be used together with the with-statement. This will
    automatically take care of logging in to Garmin Connect before any
    further interactions and logging out after the block completes or
    a failure occurs.

    Parameters
    ----------
    username
        Garmin connect username or email address
    password
        Garmin connect password
    session
        A Requests session

    Examples
    --------
    .. code-block:: python

        >>> with GarminClient("my.sample@sample.com", "secretpassword") as client:
        >>>     response = client.session.get(
        >>>         "https://connect.garmin.com/modern/{endpoint}",
        >>>         params={...},
        >>>     )

    """

    username: str = attr.ib(default=config.get("username"))
    password: str = attr.ib(
        default=config.get("password").get(), repr=False, converter=Password
    )
    session: requests.Session = attr.ib(default=None, repr=False)
    user_agent: str = attr.ib(default=config["user-agent"])

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.disconnect()

    def connect(self):
        if (not bool(self.username)) or (not bool(self.password)):
            raise ConnectionError(
                "Missing credentials. Your forgot to provide username or password. "
                f"username: '{self.username}'. password: '{self.password}'"
            )
        self.session = self.session or cloudscraper.create_scraper(
            browser={
                "browser": "firefox",
                "platform": "windows",
                "mobile": False,
            }
        )

        self._authenticate()

    def disconnect(self):
        if self.session:
            self.session.close()
            self.session = None

    def _authenticate(self):
        logger.info(f"Authenticating {self!r}")
        auth_response = self.session.post(
            ENDPOINTS["SSO_LOGIN_URL"],
            headers={
                "origin": "https://sso.garmin.com",
                "User-Agent": self.user_agent,
            },
            params={"service": "https://connect.garmin.com/modern"},
            data={
                "username": self.username,
                "password": self.password.get(),
                "embed": "false",
            },
        )
        logger.debug("got auth response: %s", auth_response.text)
        if auth_response.status_code != 200:
            raise ConnectionError(
                "authentication failure: did you enter valid credentials?"
            )
        auth_ticket_url = extract_auth_ticket_url(auth_response.text)

        logger.info("Claiming auth ticket")
        response = self.session.get(auth_ticket_url)
        if response.status_code != 200:
            raise ConnectionError(
                f"auth failure: failed to claim auth ticket: {auth_ticket_url}: {response.status_code}\n{response.text}"
            )

        # appears like we need to touch base with the old API to initiate
        # some form of legacy session. otherwise certain downloads will fail.
        self.session.get("https://connect.garmin.com/legacy/session")

    @property
    def connected(self) -> bool:
        return self.session is not None

    def get(
        self, url: str, err_message: str, tolerate: Tuple = (), params: dict = None
    ) -> requests.Response:
        """Send a get request on an authenticated session and tolerate some response codes

        Parameters
        ----------
        url
            Endpoint you want to query
        err_message
            In case of error, this message will be logged and raised with an exception
        tolerate
            Wich HTML response codes to tolerate.
        params
            URL parameters to append to the URL.
        Returns
        -------
        Response of the GET query
        """

        if not self.connected:
            raise ConnectionError(
                "Attempt to use GarminClient without being connected. Call connect() before first use."
            )

        response = self.session.get(url, params=params, headers={"NK": "NT"})
        if response.status_code in tolerate:
            return response
        elif response.status_code != 200:
            err_message += f"\nResponse code: {response.status_code}\n{response.text}"
            logger.error(err_message)
            raise ConnectionError(err_message)
        else:
            return response

    def get_activity(self, activity_id, fmt) -> requests.Response:
        """Get an activity from its ID on the requested format

        Parameters
        ----------
        activity_id
            Activity ID on Garmin Connect
        fmt
            Format you wish to download.

        Returns
        -------
        requests.Response
            Response content of the request to Garmin Connect
        """
        format_parameters = config["activities"].get(fmt)
        if not format_parameters:
            raise ValueError(
                f"Parameters for downloading the format '{fmt}' have not been provided."
            )

        response = self.get(
            url=format_parameters["endpoint"].format(id=activity_id),
            err_message=f"Failed to fetch '{fmt}' for activity id {activity_id}.",
            tolerate=tuple(format_parameters.get("tolerate", tuple())),
        )
        return response

    def list_activities(self) -> List[Dict]:
        """List all historical activities on Garmin Connect."""
        batch_size = 100
        activities = []
        for start_index in range(0, sys.maxsize, batch_size):
            response = self.get(
                url=config["endpoints"]["ACTIVITY_LIST"],
                params={"start": start_index, "limit": batch_size},
                err_message=f"Failed to fetch activities {start_index} to {start_index + batch_size - 1}.",
            )
            next_batch = json.loads(response.text)
            if not next_batch:
                break
            activities.extend(next_batch)

        return activities

    def get_wellness(self, date: pendulum.DateTime) -> requests.Response:
        """Get wellness data for a given date

        Parameters
        ----------
        date
            Date for which you want to fetch wellness data

        Returns
        -------
        requests.Response
            Response content of the request to Garmin Connect
        """

        response = self.get(
            url=config["wellness"]["endpoint"].format(date=date.format("YYYY-MM-DD")),
            err_message=f"Failed to fetch wellness data for date {date!r}.",
            tolerate=tuple(config["wellness"]["tolerate"]),
        )
        return response
